/*
 * Copyright (C) 2002-2003  Pekka Enberg <Pekka.Enberg@cs.helsinki.fi>
 *
 * Distributed under the terms of the GNU General Public License
 * version 2 or later.
 */
#include <assert.h>
#include <curses.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "ansi-esc.h"
#include "bin-file.h"
#include "colors.h"
#include "editor-context.h"
#include "edit-buffer.h"
#include "error.h"
#include "screen.h"

/* Curses type KEY_xxx check if META key was used for character.  */
#define KEY_META(ch) (0x1000 | ch)

/* ALT and ESC are both the same META key for terminals.  */
#define META_KEY_CODE 0x1B

int get_char(void)
{
	int ch = getch();
	if (ch == META_KEY_CODE) {
		/* After META comes the actual key we're interested in.  */
		ch = KEY_META(getch());
	}
	return ch;
}

static char highascii_chars[15][11] = {
	{ 218, 191, 192, 217, 196, 179, 195, 180, 193, 194, 197 }, /* single */
	{ 201, 187, 200, 188, 205, 186, 204, 185, 202, 203, 206 }, /* double horizontal */
	{ 214, 183, 211, 189, 196, 186, 199, 182, 208, 210, 215 }, /* double vertical */
	{ 213, 184, 212, 190, 205, 179, 198, 181, 207, 209, 216 }, /* double horizontal */
	{ 176, 177, 178, 219, 220, 223, 221, 222,  22, 254,  32 }, /* solid blocks */
	{ 197, 206, 215, 216,   1,   2,   3,   4,   5,   6,  32 },
	{  16,  17,  18,  19,  21,  23,  25,  29,  30,  31,  32 },
	{  28, 168, 127, 128, 129, 130, 131, 132, 133, 134,  32 },
	{ 135, 136, 137, 138, 139, 140, 141, 142, 143, 144,  32 },
	{ 145, 146, 147, 148, 149, 150, 151, 152, 153, 154,  32 },
	{ 155, 156, 157, 158, 159, 160, 161, 162, 163, 164,  32 },
	{ 165, 166, 167, 169, 170, 171, 172, 173, 174, 175,  32 },
	{ 224, 225, 226, 227, 228, 229, 230, 231, 232, 233,  32 },
	{ 234, 235, 236, 237, 238, 239, 240, 241, 242, 243,  32 },
	{ 244, 245, 246, 247, 248, 249, 250, 251, 252, 253,  32 }
};

static unsigned long selected_set = 4;

int get_printable_char(int ch)
{
	switch (ch) {
		case KEY_F(1):
			return highascii_chars[selected_set][0];
		case KEY_F(2):
			return highascii_chars[selected_set][1];
		case KEY_F(3):
			return highascii_chars[selected_set][2];
		case KEY_F(4):
			return highascii_chars[selected_set][3];
		case KEY_F(5):
			return highascii_chars[selected_set][4];
		case KEY_F(6):
			return highascii_chars[selected_set][5];
		case KEY_F(7):
			return highascii_chars[selected_set][6];
		case KEY_F(8):
			return highascii_chars[selected_set][7];
		case KEY_F(9):
			return highascii_chars[selected_set][8];
		case KEY_F(10):
			return highascii_chars[selected_set][9];
	}
	return ch;
}

/*
 * Commands issued in editor
 */

void cmd_resize(void)
{
	/* Currently we don't support resizing the terminal while
	   running.  */
	error("Terminal was resized.");
}

void cmd_move_up(struct edit_buffer *buf, struct screen *scr)
{
	if (scr->cursor_y > 0)
		scr->cursor_y--;
	else if (buf->start_y > 0)
		buf->start_y--;
}

void cmd_move_down(struct edit_buffer *buf, struct screen *scr)
{
	if (scr->cursor_y < (scr->height - 1))
		scr->cursor_y++;
	else if (buf->start_y + scr->height < buf->height)
		buf->start_y++;
}

void cmd_move_left(struct edit_buffer *buf, struct screen *scr)
{
	if (scr->cursor_x > 0)
		scr->cursor_x--;
	else if (buf->start_x > 0)
		buf->start_x--;
}

void cmd_move_right(struct edit_buffer *buf, struct screen *scr)
{
	if (scr->cursor_x < (scr->width - 1))
		scr->cursor_x++;
	else if (buf->start_x < (buf->width - scr->width))
		buf->start_x++;
}

void cmd_move_to_end(struct edit_buffer *buf, struct screen *scr)
{
	scr->cursor_x = (scr->width - 1);
	buf->start_x = buf->width - scr->width;
}

void cmd_move_to_start(struct edit_buffer *buf, struct screen *scr)
{
	scr->cursor_x = 0;
	buf->start_x = 0;
}

void cmd_print_char(struct edit_buffer *buf, struct screen *scr,
		    struct editor_context *ctx, int ch)
{
	unsigned long attr = COLOR_ATTR(ctx->fg_color, ctx->bg_color);

	edit_buffer_put(buf,
			buf->start_x + scr->cursor_x,
			buf->start_y + scr->cursor_y,
			CHAR_ATTR_TO_INT(attr, get_printable_char(ch)));
}

void cmd_move_page_up(struct edit_buffer *buf, struct screen *scr)
{
	if (buf->start_y > scr->height) {
		buf->start_y -= scr->height;
	} else {
		buf->start_y = 0;
		scr->cursor_y = 0;
	}
	screen_redraw();
}

void cmd_move_page_down(struct edit_buffer *buf, struct screen *scr)
{
	if (buf->start_y + (scr->height * 2) < buf->height) {
		buf->start_y += scr->height;
	} else {
		buf->start_y = buf->height - scr->height;
		scr->cursor_y = (scr->height - 1);
	}
	screen_redraw();
}

#define MAX_BG_COLOR 7
#define MAX_FG_COLOR 15

#define DECREASE_AND_WRAP(c, max) (c >   0 ? c - 1 : max)
#define INCREASE_AND_WRAP(c, max) (c < max ? c + 1 :   0)

void cmd_next_fg_color(struct editor_context *ctx)
{
	ctx->fg_color = INCREASE_AND_WRAP(ctx->fg_color, MAX_FG_COLOR);
}

void cmd_prev_fg_color(struct editor_context *ctx)
{
	ctx->fg_color = DECREASE_AND_WRAP(ctx->fg_color, MAX_FG_COLOR);
}

void cmd_next_bg_color(struct editor_context *ctx)
{
	ctx->bg_color = INCREASE_AND_WRAP(ctx->bg_color, MAX_BG_COLOR);
}

void cmd_prev_bg_color(struct editor_context *ctx)
{
	ctx->bg_color = DECREASE_AND_WRAP(ctx->bg_color, MAX_BG_COLOR);
}

static bool select_highascii_set(int ch)
{
#define CASE_SET(idx) \
	case KEY_META(KEY_F(idx)): \
		selected_set = idx - 1; \
		break; \

	switch (ch) {
		CASE_SET(1)
		CASE_SET(2)
		CASE_SET(3)
		CASE_SET(4)
		CASE_SET(5)
		CASE_SET(6)
		CASE_SET(7)
		CASE_SET(8)
		CASE_SET(9)
		CASE_SET(10)
		default:
			return false;
	}
	return true;
}

static bool move_cursor(int ch, struct edit_buffer * buf, struct screen * scr)
{
	switch (ch) {
		case KEY_UP:
			cmd_move_up(buf, scr);
			break;
		case KEY_DOWN:
			cmd_move_down(buf, scr);
			break;
		case KEY_LEFT:
			cmd_move_left(buf, scr);
			break;
		case KEY_RIGHT:
			cmd_move_right(buf, scr);
			break;
		case KEY_END:
			cmd_move_to_end(buf, scr);
			break;
		case KEY_HOME:
			cmd_move_to_start(buf, scr);
			break;
		case KEY_PPAGE:
			cmd_move_page_up(buf, scr);
			break;
		case KEY_NPAGE:
			cmd_move_page_down(buf, scr);
			break;
		default:
			return false;
	}
	return true;
}

static bool change_color(int ch, struct editor_context * ctx)
{
	switch (ch) {
		case KEY_META(KEY_UP):
			cmd_next_bg_color(ctx);
			break;
		case KEY_META(KEY_DOWN):
			cmd_prev_bg_color(ctx);
			break;
		case KEY_META(KEY_RIGHT):
			cmd_next_fg_color(ctx);
			break;
		case KEY_META(KEY_LEFT):
			cmd_prev_fg_color(ctx);
			break;
		default:
			return false;
	}
	return true;
}

/*
 * Main editor loop
 */

static void edit_loop(struct edit_buffer *buf)
{
	struct editor_context ctx = {
		.fg_color = 0x07,
		.bg_color = 0x00
	};

	struct screen *scr = screen_init();
	bool quit = false;

	while (!quit) {
		screen_draw_edit_buffer(scr, buf);
		screen_print_status(buf, scr, &ctx, highascii_chars[selected_set]);
		move(scr->cursor_y, scr->cursor_x);

		int ch = get_char();
		if (ch == ERR)
			error("getch() returned ERR");

		if (select_highascii_set(ch)
		    || move_cursor(ch, buf, scr)
		    || change_color(ch, &ctx))
			continue;

		switch (ch) {
			case KEY_META('x'):
			case KEY_META('X'):
				quit = true;
				break;
			case KEY_RESIZE:
				cmd_resize();
				break;
			case KEY_BACKSPACE:
				cmd_move_left(buf, scr);
				cmd_print_char(buf, scr, &ctx, ' ');
				break;
			default:
				cmd_print_char(buf, scr, &ctx, ch);
				cmd_move_right(buf, scr);
		}
	}
	screen_release(scr);
}

static void usage(char * argv[])
{
	printf("usage: %s [-c <columns> -r <rows>] [filename]\n", argv[0]);
}

int main(int argc, char *argv[])
{
	unsigned long edit_buffer_cols = 160;
	unsigned long edit_buffer_rows = 1000;

	for (;;) {
		int arg_index = getopt(argc, argv, "hc:r:");
		if (arg_index == -1) {
			break;
		}
		switch (arg_index) {
		case 'h':
			usage(argv);
			return EXIT_SUCCESS;
		case 'c':
			edit_buffer_cols = atoi(optarg);
			break;
		case 'r':
			edit_buffer_rows = atoi(optarg);
			break;
		default:
			usage(argv);
			return EXIT_FAILURE;
		}
	}

	struct edit_buffer *buf = edit_buffer_create(edit_buffer_cols, edit_buffer_rows);
	edit_buffer_clear(buf);

	if (argv[optind] != NULL) {
		FILE *input = fopen(argv[optind], "r");
		if (input) {
			if (bin_file_check(argv[optind]))
				bin_file_read(input, buf, buf->width);
			else
				ans_read(input, buf);
			fclose(input);
		}
	}
	edit_loop(buf);
	edit_buffer_release(buf);
	return EXIT_SUCCESS;
}
