/*
 * Copyright (C) 2002-2003  Pekka Enberg <penberg@iki.fi>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <assert.h>
#include <curses.h>
#include <stdlib.h>

#include "colors.h"
#include "editor-context.h"
#include "edit-buffer.h"
#include "error.h"
#include "screen.h"

static void init_curses(void)
{
	savetty();
	initscr();
	nonl();
	cbreak();
	noecho();
	keypad(stdscr, TRUE);

	if (has_colors() == FALSE) {
		error("Your terminal doesn't support colors.");
	}
	start_color();
}

static void release_curses(void)
{
	resetty();
	endwin();
}

struct screen * screen_init(void)
{
	init_curses();
	init_color_pairs();
	struct screen * ret = malloc(sizeof(struct screen));
	getmaxyx(stdscr, ret->height, ret->width);
	if (ret->height == -1 || ret->width == -1)
		error("Could not get screen dimensions from curses.");

	/* Leave a free line for the status bar.  */
	ret->height--;
	#if 0
	/* Set IBM CP437 character set.  */
	printf ("\e(U");
	#endif
	return ret;
}

void screen_release(struct screen * screen)
{
	free(screen);
	release_curses();
	#if 0
	/* Set "UNIX" character set.. */
	printf ("\e(B");
	#endif
}

static void set_attribute(int attribute)
{
	int fg_color = (attribute & 0x0F);
	int bg_color = (attribute & 0xF0) >> 4;
	if (fg_color > 7) {
		attron(A_BOLD);
		fg_color -= 8;
	}
	attron(COLOR_PAIR(attr_to_color_pair(fg_color, bg_color)));
}

static void unset_attribute(int attribute)
{
	int fg_color = (attribute & 0x0F);
	int bg_color = (attribute & 0xF0) >> 4;
	if (fg_color > 7) {
		attroff(A_BOLD);
		fg_color -= 8;
	}
	attroff(COLOR_PAIR(attr_to_color_pair(fg_color, bg_color)));
}

void screen_draw_edit_buffer(struct screen * scr, struct edit_buffer *buf)
{
	assert(buf->start_x + scr->width <= buf->width);
	assert(buf->start_y + scr->height <= buf->height);
	assert(buf->start_x >= 0);
	assert(buf->start_y >= 0);

	int x, y;

	for (y = 0; y < scr->height; y++) {
		for (x = 0; x < scr->width; x++) {
			int attribute = (edit_buffer_get(buf,
							 buf->start_x + x,
							 buf->start_y +
							 y) & 0xFF00) >> 8;
			int character = edit_buffer_get(buf,
							buf->start_x + x,
							buf->start_y +
							y) & 0xFF;
			set_attribute(attribute);
			mvprintw(y, x, "%c", character);
			unset_attribute(attribute);
		}
	}
}

#define RED_ON_BLACK 1

void screen_print_status(struct edit_buffer *buf, struct screen *scr,
			 struct editor_context *ctx, char * printable_chars)
{
	/*
	 * Last line on screen is the status bar
	 */
	move(scr->height, 0);
	clrtoeol();

	attron(COLOR_PAIR(RED_ON_BLACK));
	mvprintw(scr->height, 1, "(%2i, %2i)",
		 scr->cursor_x + buf->start_x + 1,
		 scr->cursor_y + buf->start_y + 1);
	attroff(COLOR_PAIR(RED_ON_BLACK));
	move(scr->height, 13);
	set_attribute(COLOR_ATTR(ctx->fg_color, ctx->bg_color));
	printw("Color");
	unset_attribute(COLOR_ATTR(ctx->fg_color, ctx->bg_color));

#define GREY_ON_BLACK COLOR_ATTR(7, 0)

	move(scr->height, scr->width - 42);
	set_attribute(GREY_ON_BLACK);

	int i;
	for (i = 0; i < 10; i++)
		printw(" %i=%c", i + 1, printable_chars[i]);
	unset_attribute(GREY_ON_BLACK);
}

void screen_redraw(void)
{
	redrawwin(stdscr);
}
